#!/bin/bash

modules=(sequence list-as-sequence vector-as-sequence \
  sequence-sgn sequence-sgn-simple \
  sequence-list-impl sequence-vector-impl \
  assoc-sgn assoc-list-impl hash-table \
  module1 module2)

programs=(hello-world sequence-test \
  sequence-sgn-test sequence-sgn-test2 \
  assoc-test assoc-test2 objects1 objects2 module-test1 \
  statprof-demo command-line-demo \
  orientation1 orientation2 \
  iterator-example nonpure-iterator-example \
  bits fold)

for fl in ${modules[@]} ; do
    ../../tools/compute-theme-d-pcode-deps.scm $fl.thi examples
done

for fl in ${modules[@]} ; do
    ../../tools/compute-theme-d-pcode-deps.scm $fl.thb examples
done

for fl in ${programs[@]} ; do
    ../../tools/compute-theme-d-pcode-deps.scm $fl.thp examples
done
